#!/usr/bin/env node
const chokidar = require('chokidar');
const fs = require('fs');
const cp = require('child_process');
const path = require('path');
const yaml = require('js-yaml');
const util = require('util');
const chalk = require('chalk');
const prettyjson = require('prettyjson');
const socketio = require('socket.io');
const http = require('http');
const Events = require('events');
const anymatch = require('anymatch');
const recurse = require('fs-recurse');
const md5hex = require('md5hex');

let args = process.argv.slice(2);

const EXECUTE = 'execute';
const SHOW_NBUILD = 'show-nbuild';
const INC_NBUILD = 'inc-nbuild';
const SHOW_VERSION = 'show-version';
const SHOW_HELP = 'show-help';
const LIST_FILES = 'list-files';
const NOOP = 'noop';

const CONFIG_DEFAULT_KEY = '*';

const LOCK_FILE = './frontman-lock';

const UTF8 = 'utf8';

const cs = require('./con');
const cl = require('./con');

let options = {
    showOptions: true,
    showConfig: true,
    showFiles: true,
    showStages: true,
    showChanges: true,
    showInfo: true,
    showData: true,
    showSTDIO: true,
    showLiveReload: true,
    showCommands: true,
    showCommandWorkDir: true,
    showCommandEnv: true,
    
    debounce: 1500,
    config: 'frontman.config.yaml',
    configKey: CONFIG_DEFAULT_KEY,
    action: EXECUTE,
    noInstall: false,
    noTest: false,
    modes: [],
    listFiles: "",
    env: {},
    exclude: [],
    task: false,
    once: false,

    liveReloadPort: 4001,
    liveReloadHost: "localhost",
    liveReload: false,
    liveReloadSecure: false,
    liveReloadMarkFile: './nbuild',
    liveReloadMarkURL: '/nbuild',
    liveReloadClient: "./lr-client.js",
};

let previewConfigOnly = false;
let listFilesMask = null;
let startDir = process.cwd();

if (fs.existsSync('./package.json')) {
    let packageJSON = JSON.parse(fs.readFileSync('./package.json', { encoding: UTF8 }));
    if (packageJSON.frontman) {
        options = merge(options, packageJSON.frontman);
    }
}

function setAction(action) {
    if (options.action !== EXECUTE) throw new Error(`Main action is already set.`);
    options.action = action;
}

while(args[0] && args[0][0] === '-') {
    let option = args[0].replace(/^-*/, "");
    args.shift();
    switch (option) {
        case 'show':
        case 's':
            let what = args.shift();
            switch (what) {
                case 'config': options.showConfig = true; break;
                case 'files': options.showFiles= true; break;
                case 'changes': options.showConfig = true; break;
                case 'stages': options.showStages = true; break;
                case 'stdio': options.showSTDIO = true; break;
                case 'info': options.showInfo = true; break;
                case 'data': options.showData = true; break;
                case 'live-reload': options.liveReload = true; break;
                case 'options': options.showOptions = true; break;
                case 'commands': options.showCommands = true; break;
                case 'command-dir': options.showCommandWorkDir = true; break;
                case 'command-env': options.showCommandEnv = true; break;
                case 'all':
                    options.showConfig = true;
                    options.showChanges = true;
                    options.showFiles = true;
                    options.showStages = true;
                    options.showInfo = true;
                    options.showData = true;
                    options.showOptions = true;
                    options.showLiveReload = true;
                    options.showCommands = true;
                    options.showCommandWorkDir = true;
                    options.showCommandEnv = true;
                    break;
            }
            break;
        case 'hide':
        case 'hd':
            let what2 = args.shift();
            switch (what2) {
                case 'config': options.showConfig = false; break;
                case 'files': options.showFiles= false; break;
                case 'changes': options.showConfig = false; break;
                case 'stages': options.showStages = false; break;
                case 'stdio': options.showSTDIO = false; break;
                case 'info': options.showInfo = false; break;
                case 'data': options.showData = false; break;
                case 'live-reload': options.liveReload = false; break;
                case 'options': options.showOptions = false; break;
                case 'commands': options.showCommands = false; break;
                case 'command-dir': options.showCommandWorkDir = false; break;
                case 'command-env': options.showCommandEnv = false; break;
                case 'all':
                    options.showConfig = false;
                    options.showChanges = false;
                    options.showFiles = false;
                    options.showStages = false;
                    options.showInfo = false;
                    options.showData = false;
                    options.showOptions = false;
                    options.showLiveReload = false;
                    options.showCommands = false;
                    options.showCommandWorkDir = false;
                    options.showCommandEnv = false;
                    break;
            }
            break;
        case 'force-rebuild':
            setAction(LOCK_FILE);
            break;
        case 'e':
        case 'env':
        case 'var':
            let vname = args.shift();
            if (vname.includes('=')) {
                let [k,v] = vname.split('=');
                options.env[k] = v;
            } else {
                options.env[vname] = args.shift();
            }
            break;
        case 'no-install':
        case 'ni':
        case 'fast':
            options.noInstall = true;
            break;
        case 'no-live-reload':
            options.liveReload = false;
            break;
        case 'live':
        case 'l':
            options.liveReload = true;
            break;
        case 'no-test':
            options.noTest = true;
            break;
        case 'nbuild':
            options.action = SHOW_NBUILD;
            break;
        case 'inc-nbuild':
        case 'i':
            options.action = INC_NBUILD;
            break;
        case 'help':
        case 'h':
            options.action = SHOW_HELP;
            break;
        case 'noop':
        case 'n':
            options.action = NOOP;
            break;
        case 'find':
        case 'list-files':
            options.action = LIST_FILES;
            listFilesMask = args.shift();
            break;
        case 'version':
        case 'v':
            options.action = SHOW_VERSION;
            break;
        case 'port':
        case 'p':
            options.liveReloadPort = parseInt(args.shift(), 10);
            break;
        case 'debounce':
        case 'd':
            options.debounce = parseInt(args.shift(), 10);
            break;
        case 'c':
        case 'config':
            options.config = args.shift();
            break;
        case 'nbuild-file':
        case 'b':
            options.liveReloadMarkFile = args.shift();
            break;
        case 'preview':
            previewConfigOnly = true;
            break;
        case 'once':
            options.once = args.shift();
            break;
        case 'task':
            options.task = args.shift();
            break;
        case 'exclude':
            options.exclude.push(args.shift());
            break;
        default:
            log(cs.error(`Unknown CLI options: ${cs.superStrong(option)}`, `see "frontman --help" for available options`));
            process.exit(1);
            break;
    }
}

if (args.length) {
    options.modes = [...options.modes, ...args];
}

if (options.showOptions) {
    log(`Modes:`, options.modes);
}

function readNBuildFromFile() {
    if (fs.existsSync(path.resolve(startDir, options.liveReloadMarkFile))) {
        return parseInt(fs.readFileSync(path.resolve(startDir, options.liveReloadMarkFile), UTF8), 10);
    } else {
        return 0;
    }
}

let builds = Object.assign(new Events(), {
    nbuild: readNBuildFromFile(),
    readNBuild() {
        this.setNBuild(readNBuildFromFile());
    },
    writeNBuild() {
        process.env.FRONTMAN_NBUILD = this.nbuild;
        fs.writeFileSync(path.resolve(startDir, options.liveReloadMarkFile), this.nbuild.toString(),{encoding: UTF8});
    },
    incNBuild() {
        this.setNBuild(this.nbuild + 1);
    },
    setNBuild(n) {
        if (this.nbuild !== n) {
            this.nbuild = n;
            this.writeNBuild();
            this.emit('nbuild', this.nbuild);
        }
    },
    onNBuildChanged(cb) {
        this.on('nbuild', cb);
    }
});

function resolvePath(fpath) {
    return path.resolve(startDir, fpath);
}

let liveReloadUpdate = (nbuild)=>{};

const actions = {
    [LOCK_FILE]() {
        fs.writeFileSync(resolvePath(LOCK_FILE), Math.random().toString());
        process.exit(0);
    },
    [SHOW_HELP]() {
        console.log(fs.readFileSync(path.resolve(__dirname, 'help.txt'), 'utf8'));
        process.exit(0);
    },
    [SHOW_VERSION]() {
        let selfSource = fs.readFileSync(process.argv[1], {encoding: UTF8});
        let versionHash = md5hex(selfSource, 6);
        console.log(fs.readFileSync(path.resolve(__dirname, './version'), 'utf8')+`-${versionHash}`);
        process.exit(0);
    },
    [NOOP]() {
        process.exit(0);
    },
    [SHOW_NBUILD](){
        console.log(readNBuildFromFile());
        process.exit(0);
    },
    [INC_NBUILD](){
        builds.incNBuild();
        process.exit(0);
    },
    [LIST_FILES](){
        recurse('.', (path, filename, type, cursor)=> {
            let item = path + '/' + filename;
            if (anymatch(listFilesMask, item))
                log(item, `[${type}]`);
            cursor();
        }, ()=>process.exit(0));

    },
    [EXECUTE](){


        if (options.showOptions) {
            log(`Options:\n${prettyjson.render(options)}`);
        }

        let config;

        if (typeof options.config === 'object') {
            config = options.config;
        } else {
            let configFile = resolvePath(options.config);
            if (!fs.existsSync(configFile) && fs.existsSync(`${configFile}.yaml`)) {
                configFile = `${configFile}.yaml`;
            }
            if (!fs.existsSync(configFile) && fs.existsSync(`${configFile}.js`)) {
                configFile = `${configFile}.js`;
            }
            if (!fs.existsSync(configFile)) {
                throw new Error(`Configuration not found: ${configFile}`);
            }

            if (options.config.toLowerCase().includes('.yaml')) {
                config = yaml.safeLoad(fs.readFileSync(options.config, UTF8));
            } else if (options.config.toLowerCase().includes('.json')) {
                config = JSON.parse(fs.readFileSync(options.config, UTF8));
            } else {
                config = new Function(`return ${fs.readFileSync(options.config, UTF8)};`)();
            }
        }

        if (config.options) {
            if (config.options.modes) config.options.modes = [...options.modes, config.options.modes];
            options = Object.assign(options, config.options);
            delete config.options;
        }

        if (options.liveReload) {
            logLiveReload(`Setting live-reload environment...`);
            // Set frontman environment variables for all processes
            options.env.FRONTMAN_SOCKETIO_SERVER =
                `http${options.liveReloadSecure?"s":""}://${options.liveReloadHost}:${options.liveReloadPort}`;
            options.env.FRONTMAN_SOCKETIO_PATH = `${options.liveReloadMarkURL}`;
            let clsrc = `
            (function () {
    function log(message,c) {
        console.log("%c[LiveReload]",
            "font-size: 16px; font-family: terminus, Consolas, monospace;" +
            "background-color: black; border: 2px solid slateblue; text-decoration: underline;" +
            "font-style: italic; color: cyan;" +
            "font-weight: bold;", message);
    }
    log("initializing socket.io...");
    let socketioLibScript = document.createElement('script');
    const socketIOServer = "{{{FRONTMAN_SOCKETIO_SERVER}}}";
    const socketIOPath = "{{{FRONTMAN_SOCKETIO_PATH}}}";
    socketioLibScript.setAttribute('src', socketIOServer + socketIOPath + "/socket.io.js");
    socketioLibScript.addEventListener('load', () => {
        log("socket.io initialized, connecting to server " + socketIOServer + socketIOPath);
        const socket = io(socketIOServer, {
            path: socketIOPath
        });
        socket.on('disconnect', () => {
            log("disconnected");
            socket.open();
        });
        let loadedBuild = null;
        socket.on('nbuild', (nbuild) => {
            if (loadedBuild === null) {
                log("currently loaded build #" +nbuild);
                loadedBuild = nbuild;
            } else if (loadedBuild !== nbuild) {
                log("new build #" + nbuild + "deployed: reloading");
                setTimeout( ()=>location.reload(), 400 );
            } else {
                log("same build #"+ nbuild);
            }
        });
    });
    document.head.appendChild(socketioLibScript);
})();
            `;
            clsrc = clsrc.replace(`{{{FRONTMAN_SOCKETIO_SERVER}}}`, options.env.FRONTMAN_SOCKETIO_SERVER);
            clsrc = clsrc.replace(`{{{FRONTMAN_SOCKETIO_PATH}}}`, options.env.FRONTMAN_SOCKETIO_PATH);
            let targetClientFile = path.resolve(startDir, options.liveReloadClient);
            fs.writeFileSync(targetClientFile, clsrc, {
                encoding: UTF8
            });
            logLiveReload(`Written live-reload client code to: ${chalk.white(targetClientFile)}`);

            logLiveReload(`Starting live reload server on port ${options.liveReloadPort} at path "${options.liveReloadMarkURL}"`);
            let liveReloadServer = socketio(options.liveReloadPort, {
                path: options.liveReloadMarkURL
            });
            let liveReloadClients = new Set();

            liveReloadUpdate = (nbuild) => {
                liveReloadClients.forEach(socket=>socket.emit('nbuild', nbuild));
            };

            liveReloadServer.on('connection', function (socket) {
                logLiveReload(`LiveReload: client connected [${socket.id}]`);
                liveReloadClients.add(socket);
                liveReloadUpdate(builds.nbuild);
                socket.on('disconnecting', () => {
                    logLiveReload(`LiveReload: client disconnected [${socket.id}]`);
                    liveReloadClients.delete(socket);
                });

                socket.on('disconnect', function () {
                    logLiveReload(`LiveReload: client disconnected [${socket.id}]`);
                    liveReloadClients.delete(socket);
                });
            });
        }

        executeConfig(config);
    }
};


//##############################################################################################################
//#   END OF MAIN PROGRAM CODE
//##############################################################################################################

function log(...args) {
    cl(...args);
}

function logLiveReload(...args) {
    if (options.showLiveReload) {
        log(chalk.bgGreen.black.underline("LIVE RELOAD"), ...args);
    }
}

function logConfig(config) {
    if (options.showConfig) {
        let result = `${cs.reportTitle('CONFIG')}`;
        for (let task of Object.keys(config)) {
            result += `\n${cs.reportTitle('TASK')} ${cs.report(task)}\n`;
            result += prettyjson.render(config[task], {});
        }
        log(result);
    }
}

function logConfigTask(name, config) {
    if (options.showConfig) {
        let result;
        if (typeof config !== 'object') {
            if (!config)
                config = "DISABLED/EMPTY";
            else if (typeof config === 'string')
                if (config[0] === '!')
                    config = config.substr(1);
                else
                    config = `DISABLED BY MODE ${chalk.bold(chalk.italic(config))}`;
            else
                config = `INCORRECT: ${config}`;
            result = `${cs.reportTitle('TASK-MODE')} ${cs.report(chalk.bold(name))} ${chalk.red(config)}\n`;
        } else {
            result = `${cs.reportTitle('TASK-MODE')} ${cs.report(name)}\n`;
            result += prettyjson.render(config, {});
        }
        log(result);
    }
}

function getProp(object, possibleKeys, deflt, transform = x=>x) {
    let result = deflt;
    for (key of possibleKeys) {
        if (key in object) {
            result = transform(object[key]);
            break;
        }
    }
    object[possibleKeys[0]] = result;
    return result;
}

const WorkDirKeys = ['workDir', 'at', 'dir', 'wdir', 'cwd','pwd'];
function getWorkDirProp(object, root = startDir) {
    return getProp(object, WorkDirKeys, root, (x)=>path.resolve(root, x));
}

const EnvKeys = ['env', 'set'];
function getEnvProp(object, inherit = options.env) {
    return getProp(object, EnvKeys, inherit, x=>Object.assign({}, inherit, x));
}

const EnabledFalseValues = ['disabled', 'off', 'false', 'no', '-', 'disable'];
const EnabledKeys = ['enabled', 'enable'];
function getEnabledProp(object) {
    let firstKey = EnabledKeys[0];
    let key = firstKey;
    let result = true;
    for (key of EnabledKeys) {
        if (key in object) break;
    }
    if (key in object) {
        object[firstKey] = object[key];
    }
    if (firstKey in object) {
        result = !EnabledFalseValues.includes(object[firstKey]);
    }
    object[firstKey] = result;
    return result;
}

function getBooleanValue(value, deflt = false) {
    return (typeof value === 'undefined') ? deflt : EnabledFalseValues.includes(value) ? false : !!value;
}

function macroCommands(cmdlist) {

    let macroses = {
        "$pull": ["git pull origin"],
        "$clean": ["rm package-lock.json", "rm -rf node_modules"],
        "$install": ["npm install"],
        "$reinstall": ["rm package-lock.json", "rm -rf node_modules", "npm install"]
    };
    let cmdlistOptions = {
        cmds: [
            {
                cmd: "frontman --noop"
            }
        ]
    };

    if (cmdlist) {
        if (cmdlist instanceof Array) {
            cmdlistOptions.cmds = cmdlist;
        } else {
            if (typeof cmdlist === 'string') {
                cmdlistOptions.cmds = [ cmdlist ];
            } else {
                cmdlistOptions = cmdlist;
            }
        }

        let result = [];
        cmdlistOptions.cmds.forEach((item) => {
            if (typeof item === 'string') {
                if (item in macroses) {
                    result.push(...macroses[item].map(macroCmd=>{ return {cmd: macroCmd};}));
                } else {
                    result.push({cmd: item});
                }
            } else {
                result.push(item);
            }

        });
        cmdlistOptions.cmds = result;

        getWorkDirProp(cmdlistOptions);
        getEnvProp(cmdlistOptions);
        getEnabledProp(cmdlistOptions);

        cmdlistOptions.cmds.forEach(cmdObject=>{
            if (cmdObject.run) cmdObject.cmd = cmdObject.run;
            getWorkDirProp(cmdObject, cmdlistOptions.workDir);
            getEnvProp(cmdObject, cmdlistOptions.env);
        });
    }

    return cmdlistOptions;
}

function macroArguments(argList) {
    let result = [];

    if (argList instanceof Array)
        result = [...argList];
    else {
        if (argList) {
            if (argList instanceof Object) {

                let dir = getWorkDirProp(argList);
                let env = getEnvProp(argList);
                let enabled = getEnabledProp(argList);
                let silent = getBooleanValue(argList.silent);
                let waitExit = parseInt(argList.waitExit, 10) || 0;

                for (let k in argList) {
                    if (k[0] === '$') {
                        env[k.substr(1)] = argList[k];
                    }
                }
                if (!argList.cmd) throw new Error(`cmd must be specified`);

                if (argList.cmd instanceof Array)
                    result = argList.cmd;
                else if (typeof argList.cmd === 'string')
                    result = argList.cmd.split(' ').filter(part => !!part);
                else
                    throw new Error(`Incorrect 'cmd' value`);

                result.workDir = dir;
                result.env = env;
                result.enabled = enabled;
                result.waitExit = waitExit;
                result.silent = silent;
            } else { //string
                result = argList.split(' ').filter(part => !!part);
            }
        } else { // !argList
            result = ["frontman", "--noop"];
        }
    }

    if (!result.length) {
        throw new Error(`void command`)
    }

    getWorkDirProp(result);
    getEnvProp(result);
    getEnabledProp(result);
    result.waitExit = result.waitExit || 0;
    result.silent = result.silent || false;
    return result;
}

let tasks = {};

function merge(target, addon, rcStack = []) {
    target = target || {};
    addon = addon || ((target instanceof Array) ? [] : {});
    if (rcStack.includes(target)) {
        throw new Error(`Recursion.`);
    }
    if (target instanceof Array && addon instanceof Array) {
        target.push(...addon);
    } else if (target instanceof Object && addon instanceof Object) {
        Object.keys(addon).forEach(key => {
            if ((addon[key] === undefined) || (addon[key] === 'DELETE')) {
                Reflect.deleteProperty(target, key);
            } else {
                if (key in target) {
                    if (target[key] instanceof Object) {
                        if (addon[key]) {
                            merge(target[key], addon[key], [...rcStack, target]);
                        } else if (addon[key] === false) {
                            delete target[key];
                        }
                    } else {
                        target[key] = addon[key];
                    }
                } else {
                    target[key] = addon[key];
                }
            }
        });
    } else {
        throw new Error(`Incompatible types to merge.`);
    }
    return target;
}

function applyModes(taskConf) {
    let modeText = "";
    if (taskConf.modes) {
        let taskModes = taskConf.modes;
        delete taskConf.modes;

        for (let mode of options.modes) {
            if (mode in taskModes) {
                switch (typeof taskModes[mode]) {
                    case 'object':
                        modeText = (modeText ? `${modeText},` : ":")  + mode;
                        taskConf = merge(taskConf, taskModes[mode]);
                        break;
                    default:
                        return mode;
                }
            }
        }
    }
    taskConf.mode = modeText;
    return taskConf;
}

function executeConfig(config) {
    if (options.configKey && options.configKey !== CONFIG_DEFAULT_KEY) {
        config = config[options.configKey];
    }

    logConfig(config);

    function resolveTaskConfig(name, originalTaskConfig) {
        let taskConfig = Object.assign({}, originalTaskConfig);
        taskConfig.name = name;
        taskConfig.install = macroCommands(taskConfig.install);
        taskConfig.test = macroCommands(taskConfig.test);
        taskConfig.build = macroArguments(taskConfig.build);
        taskConfig.run = macroArguments(taskConfig.run);
        taskConfig.nbuild = getBooleanValue(taskConfig.nbuild, true);
        taskConfig.liveReload = getBooleanValue(taskConfig.liveReload, true);

        let watchConf = taskConfig.watch || ["**/*"];
        let ignoreConf = taskConfig.ignore || taskConfig.ignored || [".git", "node_modules", "bower_components"];
        let watchFiles;
        let watchOptions = {
            ignored: ignoreConf
        };

        delete taskConfig.watch;
        delete taskConfig.ignore;
        delete taskConfig.ignored;

        if (watchConf instanceof Array) {
            watchFiles = watchConf;
        } else if (watchConf instanceof Object) {
            watchFiles = watchConf.files;
            delete watchConf.files;
            watchOptions = Object.assign(watchOptions, watchConf);
        } else if (typeof watchConf === 'string') {
            watchFiles = [ watchConf ];
        } else {
            throw new Error(`No correct watch specified in ${options.config} -> ${name}`);
        }

        watchOptions.cwd = watchOptions.cwd ? resolvePath(watchOptions.cwd) : startDir;
        if (watchOptions.awaitWriteFinish || watchOptions.wait || watchOptions.poll) {
            if (!(watchOptions.awaitWriteFinish instanceof Object)) {
                watchOptions.awaitWriteFinish = {
                    stabilityThreshold: 2000,
                    pollInterval: 200
                }
            }
            if (watchOptions.wait) {
                if (typeof watchOptions.wait === 'number')
                    watchOptions.awaitWriteFinish.stabilityThreshold = watchOptions.wait;

            }
            if (watchOptions.poll) {
                watchOptions.awaitWriteFinish.pollInterval = watchOptions.poll;
            }
        }
        if (watchOptions.wait === false) {
            watchOptions.awaitWriteFinish = false;
        }
        delete watchOptions.poll;
        delete watchOptions.wait;

        watchOptions.ignoreInitial = true;

        taskConfig.watchOptions = watchOptions;
        taskConfig.watchFiles = [...watchFiles, LOCK_FILE];
        return taskConfig;
    }

    const State = {
        StartingUp: "Start",
        WaitingForChanges: 'Watch',
        WaitingForDebounce: "Debounce",
        Installing: "Install",
        StartingWatcher: "InitWatcher",
        RebuildAndRun: "RebuildAndRun",
        Building: "Build",
        Testing: "Test",
        Running: "Run",
        WaitingForPreviousProcessToStop: "Stop",
        AllFinished: "AllFinished"
    };

    let tasksRunning = 0;

    for (let taskName of Object.keys(config)) {
        if (options.exclude.includes(taskName)) {
            logConfigTask(taskName, `!EXCLUDED BY COMMAND-LINE`);
            continue;
        }
        if (options.task && options.task !== taskName) {
            logConfigTask(taskName, `!ALL EXCLUDED EXCEPT [${chalk.underline(options.task)}]`);
            continue;
        }
        let modesApplied = applyModes(config[taskName]);
        if (typeof modesApplied === 'string' || !modesApplied) {
            logConfigTask(taskName, modesApplied);
            continue;
        }
        let taskConfig = resolveTaskConfig(taskName, modesApplied);
        logConfigTask(taskName, taskConfig);
        let task = {
            name: taskName,
            state: State.StartingUp,
            config: taskConfig,
            buildProcess: null,
            runProcess: null,
            stopBuild: false,
            stopRun: false,
            changes: {},
            previousChanges: {},
            mode: taskConfig.mode,
            debounceTimeout: null,
            resetChanges() {
                this.previousChanges = this.changes;
                this.changes = {};
            },
            restorePreviousChanges() {
                this.changes = Object.assign(this.changes, this.previousChanges);
            },
            addChange(event, file) {
                this.logChange(event, file);
                this.changes[file] = event;
            },
            logRelevantChanges() {
                let report = "";
                for (let k of Object.keys(this.changes)) {
                    report += `${chalk.bgCyan.black.underline(this.changes[k])} ${chalk.italic.cyan(k)}\n`;
                }
                this.notify(`Relevant changes:`);
                this.log("\n",report);
            },
            log(...args) {
                log(chalk.bgCyan.bold.black(`[${chalk.underline(this.name)}${this.mode}]`), ...args);
            },
            notify(message) {
                this.log(chalk.cyan.dim(message));
            },
            logState(state, data = null) {
                this.log(chalk.cyan.italic.bold(state));
                if (data) this.logData(data);
            },
            prettydata(data, inlineArrays = true) {
                return prettyjson.render(data, {
                    inlineArrays: inlineArrays,
                    emptyArrayMs: chalk.dim.bgCyan.black("<empty>"),
                    noColor: false,
                    keysColor: "blue",
                    dashColor: "cyan",
                    numberColor: "magenta",
                    stringColor: "green",

                    defaultIndentation: 2,
                } , 2);
            },
            logData(data) {
                this.log("\n", this.prettydata(data));
            },
            logCommand(command) {
                let toLog = (command instanceof Array) ? command.join(' ') :
                    ((command instanceof Object) ? JSON.stringify(command.cmd) : command);
                if (command.workDir && options.showCommandWorkDir) {
                    toLog += `\n  @${command.workDir}`;
                }
                if (command.env && options.showCommandEnv) {
                    this.log(chalk.bgYellow.red.underline('[ >>> ]'), chalk.yellow.italic(toLog),
                        `\n${prettyjson.render(command.env)}`);
                } else {
                    this.log(chalk.bgYellow.red.underline('[ >>> ]'), chalk.yellow.italic(toLog));
                }
            },
            logError(error) {
                let toLog = (error instanceof Error) ? error.stack : JSON.stringify(error);
                this.log(chalk.bgRed.bold.black("ERROR!"), chalk.red.underline(toLog));
            },
            logSuccess(message) {
                this.log(chalk.bgGreen.bold.black(message));
            },
            logCritical(error) {
                let toLog = (error instanceof Error) ? error.stack : JSON.stringify(error);
                this.log(chalk.bgRed.bold.black.italic("CRITICAL ERROR!"), chalk.red.bold.underline(toLog));
            },
            logFatal(error) {
                let toLog = (error instanceof Error) ? error.stack : JSON.stringify(error);
                this.log(chalk.bgRed.bold.yellow.italic("FATAL ERROR!"), chalk.bgYellow.red.bold.underline(toLog));
                process.exit(1);
            },
            logChange(event, file) {
                this.log(chalk.green.underline(`${event}:`), chalk.cyan(file));
            },
            start() {
                if (!previewConfigOnly) {
                    this.logState('Starting...');
                    tasksRunning++;
                    this.buildProcess = null;
                    this.runProcess = null;
                    this.stopBuild = false;
                    this.stopRun = false;

                    try {
                        this.install();
                    } catch (e) {
                        tasksRunning--;
                        if (!tasksRunning) process.exit(-1);
                        throw e;
                    }
                    if (options.once) {
                        this.rebuildAndRun();
                    } else {
                        this.watch();
                    }
                    this.handleChange("INIT", "*");
                } else {
                    if (options.showOptions)
                        this.notify(`Preview Mode. Will not run, only display configuration.`);
                    this.logData(this.config);
                }
            },
            install() {
                this.logState(State.Installing);
                if (!options.noInstall) {
                    let seq = this.config.install;
                    this.runCommandSequence(seq);
                } else {
                    if (options.showOptions)
                        this.notify(`Install disabled by options.`);
                }

            },
            test() {
                this.logState(State.Testing);
                if (!options.noTest) {
                    let seq = this.config.test;
                    this.runCommandSequence(seq);
                } else {
                    if (options.showOptions)
                        this.notify(`Test disabled by options.`);
                }
            },
            watch() {
                this.logState(State.StartingWatcher);
                this.watcher = chokidar.watch(this.config.watchFiles, this.config.watchOptions);
                this.watcher.on('all', this.handleChange);
            },
            handleChange(event, file) {
                this.addChange(event, file);
                if (this.debounceTimeout) {
                    clearTimeout(this.debounceTimeout);
                    this.debounceTimeout = null;
                }
                //this.logState(State.WaitingForDebounce);
                this.debounceTimeout = setTimeout(() => {
                    this.debounceTimeout = null;
                    this.rebuildAndRun();
                }, options.debounce);
            },
            rebuildAndRun() {
                this.logState(State.RebuildAndRun);
                if (this.debounceTimeout) {
                    clearTimeout(this.debounceTimeout);
                    this.debounceTimeout = null;
                }

                if (this.stopBuild) {
                    this.logError(new Error(`Expecting for previous build to terminate, canceling rebuild and run.`));
                    return;
                }

                if (this.stopRun) {
                    this.logError(new Error(`Expecting for previous run to terminate, canceling rebuild and run.`));
                    return;
                }

                if (this.buildProcess) {
                    this.stopBuild = true;
                    this.logState(State.WaitingForPreviousProcessToStop);
                    this.buildProcess.stdin && this.buildProcess.stdin.write("\x03\x03");
                    this.buildProcess.kill();
                } else if (this.runProcess) {
                    this.stopRun = true;
                    this.logState(State.WaitingForPreviousProcessToStop);
                    this.runProcess.stdin && this.runProcess.stdin.write("\x03\x03");
                    this.runProcess.kill();
                } else {
                    this.build().then((code) => {
                        if (this.stopBuild) {
                            this.stopBuild = false;
                            if (options.once) {
                                process.exit(-1);
                            } else
                                this.rebuildAndRun();
                        } else if (code === 0) {
                            if (this.config.nbuild) {
                                builds.incNBuild();
                                this.logSuccess(`Build #${builds.nbuild} successfull!`);
                            } else {
                                this.logSuccess(`Build successfull!`);
                            }

                            this.test();
                            let runPromise = this.run();
                            if (this.config.liveReload) {
                                liveReloadUpdate(builds.nbuild);
                            }
                            runPromise.then((code) => {
                                if (this.stopRun) {
                                    this.notify(`Terminated previous run, queueing a new one...`);
                                    this.stopRun = false;
                                    if (options.once) {
                                        process.exit(-1);
                                    } else
                                        this.rebuildAndRun();
                                } else if (code === 0) {
                                    this.logSuccess(`Task completed successfully`);
                                    if (options.once) {
                                        this.logState(State.AllFinished);
                                        tasksRunning--;
                                        if (!tasksRunning) process.exit(0);
                                    } else {
                                        this.logState(State.WaitingForChanges);
                                    }

                                } else {
                                    this.logError(`Run failed with code ${code}.`);
                                    this.restorePreviousChanges();
                                    this.logState(State.WaitingForChanges);
                                }
                            });
                        }
                    });
                }
            },
            build() {
                this.logState(State.Building);
                this.logRelevantChanges();
                return this.spawn(this.config.build, 'buildProcess');
            },
            run() {
                this.logState(State.Running);
                this.logRelevantChanges();
                this.resetChanges();
                return this.spawn(this.config.run, "runProcess");
            },
            runCommandSequence(seq) {
                let {cmds, enabled} = seq;
                if (enabled) {
                    for (let {cmd, strict, workDir, env, silent} of cmds) {
                        try {
                            this.logCommand(cmd);
                            cp.execSync(cmd, {
                                stdio: options.showSTDIO ? (silent ? undefined : "inherit") : undefined,
                                cwd: workDir || startDir,
                                env: Object.assign({}, process.env, env)
                            });
                        } catch (error) {
                            if (strict) {
                                this.logCritical(error);
                                return false;
                            } else {
                                this.logError(error);
                            }
                        }
                    }
                }

                return true;
            },
            spawn(command, procName) {
                return new Promise((resolve) => {
                    let [prog, ...args] = command;
                    let {enabled, env, workDir, waitExit,silent} = command;
                    let proc;
                    if (enabled) {
                        this.logCommand(command);
                        proc = cp.spawn(prog, args, {
                            cwd: workDir,
                            stdio: (options.showSTDIO && !silent) ? "inherit" : undefined,
                            env: Object.assign({}, process.env, env)
                        });
                        this[procName] = proc;
                        proc.on('error', this.logCritical);
                        proc.once('exit', (code) => {
                            this[procName] = null;
                            if (code) {
                                this.logError(new Error(`${procName} failed!`));
                            }
                            setTimeout(() => {
                                resolve(code);
                            }, waitExit);

                        });
                    } else {
                        this[procName] = null;
                        resolve(0);
                    }
                });
            }
        };
        for (let k of Object.keys(task)) {
            if (typeof task[k] === 'function') task[k] = task[k].bind(task);
        }
        tasks[taskName] = task;
        task.start();
    }
}

actions[options.action]();
